import abc
from datetime import datetime
from enum import IntEnum
from sys import stderr


class Logger:
    """
    Logger class provides the logging funcionality. It maintains a level, and prints out a message if and
    only if the logger level is less than or equal to the level provided as an argument to the print method.
    It also maintains a list of printers and prints each message by all the printers.
    """

    class Level(IntEnum):
        """
        Logger levels to ease the usage.
        The base class in IntEnum, since Enum does not support comparing values, which can be anything.
        It is possible to implement the comparison for ordinary Enum, but then again -- why?
        """
        SEVERE = 0
        WARNING = 1
        INFO = 2
        FINE = 3
        FINER = 4
        FINEST = 5

    def __init__(self, name):
        self.printers = list()
        self.name = name
        self.level = self.Level.SEVERE
        self.formatter = None

    def set_level(self, level):
        self.level = level

    def log(self, level, message):
        if self.level >= level:
            if self.formatter is not None:
                message = self.formatter.format(self, message)

            for printer in self.printers:
                printer.print(message)

    def add_printer(self, printer):
        self.printers.append(printer)

    def set_formatter(self, formatter):
        self.formatter = formatter


class Printer(abc.ABC):
    """
    Abstract class for printers. It defines just the print method.
    """

    @abc.abstractmethod
    def print(self, message):
        """
        Prints the message
        :param message: The message to be printed out
        :return: nothing
        """
        pass


class BasicPrinter(Printer):
    """
    The basic printer, just prints messages to standard output
    """

    def print(self, message):
        print(message)


class FilePrinter(Printer):
    """
    Prints a timestamp in front of each message, prints to stdout
    """

    def __init__(self, filename):
        self.filename = filename

    def print(self, message):
        with open(self.filename, mode='at') as f:
            print(message, file=f)


class ErrorPrinter(Printer):
    """
    Prints messages to stderr
    """

    def print(self, message):
        print(message, file=stderr)


class Formatter(abc.ABC):
    """
    Abstract class for formatters.
    """

    @abc.abstractmethod
    def format(self, logger, message):
        """
        Formats the given message
        :param logger: The logger whose properties can be used for formatting
        :param message: The message to be formatted
        :return: The formatted message
        """
        pass


class NameFormatter(Formatter):
    """
    Extends the message by the name of the logger
    """

    def format(self, logger, message):
        return logger.name + ': ' + message


class TimestampFormatter(Formatter):
    """
    Extends the message by a timestamp
    """

    def format(self, logger, message):
        return str(datetime.now()) + ': ' + message


logger = Logger("The Logger")
logger.add_printer(BasicPrinter())
logger.add_printer(FilePrinter('../output.log'))
logger.set_level(Logger.Level.INFO)
logger.set_formatter(TimestampFormatter())
logger.log(Logger.Level.INFO, "Logged message")
logger.log(Logger.Level.FINE, "Message ignored due to loglevel")
