import abc
import os
import shutil
from os import listdir


class Shell:
    """
    Extensible shell, allows for execution of commands implementing the Command interface
    """

    def __init__(self):
        """
        Initializer of the class
        """
        self.commands = dict()

    def add_command(self, command):
        self.commands[command.name()] = command

    def run(self):
        cmd = input('Your command, sir: ')
        cmd = cmd.strip()
        while cmd != 'exit':
            if cmd == 'help':
                print('List of supported commands:')
                print('===========================')
                print()
                for name, command in self.commands.items():
                    print(name + ': ', end='')
                    print(command.help())
            else:
                parts = cmd.split(' ')
                if parts[0] in self.commands.keys():
                    self.commands[parts[0]].exec(parts[1:])
                else:
                    print('Command not found.')

            cmd = input('Your command, sir: ')
            cmd = cmd.strip()


class Command(abc.ABC):
    @abc.abstractmethod
    def name(self):
        """
        :return: The name of the command, i.e., its string representation as given at the stdin
        """
        pass

    @abc.abstractmethod
    def help(self):
        """
        :return: The string with help to the command.
        """
        pass

    @abc.abstractmethod
    def exec(self, args):
        """
        Executes the command. It prints out any potential error messages.
        :param args: The arguments of the command, such as filenames to be copied,
        or directory to be switched to
        :return: Nothing
        """
        pass


class CommandCd(Command):
    """
    Changes the current directory.
    """

    def exec(self, args):
        if len(args) == 1:
            try:
                os.chdir(args[0])
            except FileNotFoundError:
                print('The specified dir does not exist.')
        else:
            print('New dir expected.')

    def help(self):
        text = '''Changes the current directory to the one given as parameter.
        If the specified directory does not exists, it does nothing.
        '''
        return text

    def name(self):
        return "cd"


class CommandLs(Command):
    """
    Prints out the content of the current directory.
    """

    def name(self):
        return "ls"

    def help(self):
        text = '''Prints out the content of the current directory, 
        i.e., the list of files and directories.
        '''
        return text

    def exec(self, args):
        if len(args) == 0:
            for filename in sorted([f for f in listdir('.')]):
                print(filename)
        else:
            print('No arguments required.')


class CommandPwd(Command):
    """
    Prints out the current directory, i.e., the path to it.
    """

    def name(self):
        return 'pwd'

    def help(self):
        text = '''Prints out the content directory. 
        '''
        return text

    def exec(self, args):
        print(os.getcwd())


class CommandCp(Command):
    """
    Copies the file given as the first argument to the file given as the second argument.
    """

    def name(self):
        return 'cp'

    def help(self):
        text = '''Copies the file given as the first parameter to the file given as the second one.
        Prints out the error description if one occurs. 
        '''
        return text

    def exec(self, args):
        if len(args) == 2:
            try:
                shutil.copy(args[0], args[1])
            except FileNotFoundError:
                print('Error copying file. Does the source exist? Does the destination dir exist?')
            except shutil.SameFileError:
                print('Source and destination file cannot be the same.')
        else:
            print('Source and target files expected.')


shell = Shell()
shell.add_command(CommandCd())
shell.add_command(CommandCp())
shell.add_command(CommandLs())
shell.add_command(CommandPwd())

shell.run()
