import abc
import sys


class TextProcessor:
    """
    Extensible text processor.
    It reads input from the stdin and writes to the stdout
    It takes a list of command line arguments with names of plugins to be applied on the text
    """

    def __init__(self, args):
        """
        Create an instance of a plugin specified by its name at command line and adds the instance to the plugin list
        :param args: The list of plugin names
        """
        self.plugins = list()
        for arg in args:
            for plugin in Plugin.available_plugins:
                if plugin.name() == arg:
                    self.plugins.append(plugin)

    def apply_plugins(self, line):
        """
        Applies all the plugins in the given order
        :param line: The line to be processed, including the newline character
        :return: The line processed by all plugins
        """
        for plugin in self.plugins:
            line = plugin.process(line)

        return line

    def process_input(self):
        """
        The 'main' function of the TextProcessor
        :return: nothing
        """
        for line in sys.stdin.readlines():
            print(self.apply_plugins(line), end='')


class Plugin(abc.ABC):
    """
    Interface for text processing plugins
    """

    @abc.abstractmethod
    def process(self, line):
        """
        Applies the plugin to the given line (string)
        :param line: The line to be processed
        :return: The processed line
        """
        pass

    @abc.abstractmethod
    def name(self):
        """
        The name of the plugin
        :return: The name of the plugin
        """
        pass

    available_plugins = list()  # list of available plugins


class ToUpper(Plugin):
    """
    Plugin to convert the text to upper case.
    """

    def name(self):
        return 'to_upper'

    def process(self, line):
        return line.upper()


class ToLower(Plugin):
    """
    Plugin to convert the text to lower case.
    """

    def name(self):
        return 'to_lower'

    def process(self, line):
        return line.lower()


class Capitalize(Plugin):
    """
    Plugin to capitalize the text.
    """

    def name(self):
        return 'capitalize'

    def process(self, line):
        words = list()

        for word in line.split():
            words.append(word.capitalize())

        return ' '.join(words) + '\n'


Plugin.available_plugins.append(ToUpper())
Plugin.available_plugins.append(ToLower())
Plugin.available_plugins.append(Capitalize())

TextProcessor(sys.argv[1:]).process_input()
